/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte, Charles Souillard - BonitaSoft S.A.
 **/

package org.ow2.bonita.facade.def.element.impl;

/**
 * Utility class to create a Deployment object containing the the XPDL file and optionally its
 */
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.facade.def.element.BusinessArchive;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.ClassDataTool;
import org.ow2.bonita.util.Misc;

/**
 * This class implements a deployment object. This object can be provided to one of the method of the ManagementAPI
 * {@link org.ow2.bonita.facade.ManagementAPI#deploy(BusinessArchive) method} to allow the deployment of a XPDL file
 * and optionally its depending java classes for hooks, performer assignments, mappers...
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class BusinessArchiveImpl implements Serializable, BusinessArchive {

  /**
   * 
   */
  private static final long serialVersionUID = 218650935088396315L;

  private static final Logger LOG = Logger.getLogger(BusinessArchiveImpl.class.getName());
  public static final String PROCESS_RESOURCE_NAME = "Process.def";

  protected Set<Resource> resources = new HashSet<Resource>(); 

  protected BusinessArchiveImpl() { }

  public BusinessArchiveImpl(ProcessDefinition clientProcess, Map<String, byte[]> resources, final Class< ? >... classes) throws IOException, ClassNotFoundException {
    Misc.checkArgsNotNull(clientProcess);

    addResource(PROCESS_RESOURCE_NAME, Misc.serialize(clientProcess));

    if (resources != null) {
      for (Map.Entry<String, byte[]> resource : resources.entrySet()) {
        addResource(resource.getKey(), resource.getValue());
      } 
    }
    if (classes != null && classes.length > 0) {
      for (Class< ? > clazz : classes) {
        if (clazz != null) {
          addResource(clazz.getName() + ".class", ClassDataTool.getClassData(clazz));
        }
      } 
    }
  }

  public BusinessArchiveImpl(Map<String, byte[]> resources) throws IOException {
    Misc.checkArgsNotNull(resources);
    for (Map.Entry<String, byte[]> resource : resources.entrySet()) {
      addResource(resource.getKey(), resource.getValue());
    }

  }

  public BusinessArchiveImpl(BusinessArchive businessArchive) {
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Building a Deployment object from businessArchive: " + resources);
    }
    for (Map.Entry<String, byte[]> entry : businessArchive.getResources().entrySet()) {
      addResource(entry.getKey(), entry.getValue());
    }
  }

  public ProcessDefinition getProcessDefinition() {
    byte[] processBytes = getResource(PROCESS_RESOURCE_NAME);
    if (processBytes == null) {
      return null;
    }
    try {
      return (ProcessDefinition) Misc.deserialize(processBytes);
    } catch (IOException e) {
      throw new BonitaRuntimeException(e);
    } catch (ClassNotFoundException e) {
      throw new BonitaRuntimeException(e);
    }
  }

  public Map<String, byte[]> getClasses() {
    return getResources(".*\\.class");
  }

  public Map<String, byte[]> getAllResources() {
    Map<String, byte[]> allResources = getResources();

    Map<String, byte[]> jarFiles = getJarFiles();
    for (Map.Entry<String, byte[]> entry : jarFiles.entrySet()) {
      String jarFile = entry.getKey();
      byte[] jar = entry.getValue();
      JarEntry jarEntry = null;
      try {
        JarInputStream jis = new JarInputStream(new ByteArrayInputStream(jar));
        while ((jarEntry = jis.getNextJarEntry()) != null) {
          String jarEntryName = jarEntry.getName(); 
          final ByteArrayOutputStream baos = new ByteArrayOutputStream();
          int c;
          final byte[] buffer = new byte[512];
          while ((c = jis.read(buffer)) != -1) {
            baos.write(buffer, 0, c);
          }
          baos.flush();
          allResources.put(jarEntryName, baos.toByteArray());
          Misc.close(baos);
        }
      } catch (IOException e) {
        throw new BonitaRuntimeException("Unable to load class: " + jarEntry.getName() + " from jar file: " + jarFile);
      }
    }
    return allResources;
  }

  public Map<String, byte[]> getJarFiles() {
    return getResources(".*\\.jar");
  }

  public byte[] getResource(String resourcePath) {
    for (Resource resource : resources) {
      if (resource.getPath().equals(resourcePath)) {
        return resource.getData();
      }
    }
    return null;
  }

  public Map<String, byte[]> getResources(String regex) {
    Map<String, byte[]> result = new HashMap<String, byte[]>();
    for (Resource resource : resources) {
      if (resource.getPath().matches(regex)) {
        result.put(resource.getPath(), resource.getData());
      }
    }
    return result;
  }

  public void addResource(String resourcePath, byte[] resourceData) {
    Misc.checkArgsNotNull(resourcePath, resourceData);
    this.resources.add(new Resource(resourcePath, resourceData));
  }

  public Map<String, byte[]> getResources() {
    Map<String, byte[]> allResources = new HashMap<String, byte[]>();
    for (Resource resource : resources) {
      allResources.put(resource.getPath(), resource.getData());
    }
    return allResources;
  }

  protected static class Resource implements Serializable {

    private static final long serialVersionUID = 4985009903197019706L;
    protected String path;
    protected byte[] data;

    protected Resource() {}

    public Resource(String path, byte[] data) {
      super();
      this.path = path;
      this.data = data;
    }
    public String getPath() {
      return path;
    }
    public byte[] getData() {
      return data;
    }
  }
}
