/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.facade.def.majorElement;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.FilterDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.MultiInstantiationDefinition;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;

/**
 * This interface represents the Activity definition.
 * It's derived from the Process Activity Definition of XPDL.
 *
 */
public interface ActivityDefinition extends ProcessElement, Serializable {

  /**
   * Join types supported by Bonita.
   */
  public static enum JoinType {
    AND, XOR
  }

  /**
   * Split types supported by Bonita.
   */
  public static enum SplitType {
    AND, XOR
  }
  
  /**
   * Returns the UUID for the ActivityDefinition.
   */
  ActivityDefinitionUUID getUUID();

  /**
   * Activity can be defined either with Route or Implementation or BlockActivity element.
   * @return true if the activity has no associated group and no associated filter.
   *
   */
  boolean isAutomatic();
  
  /**
   * Returns the SubFlowDefinition interface if this activity has subflow implementation
   * otherwise returns null.
   */
  String getSubflowProcessName();

  /**
   * Returns the performer of the activity.
   * For an activity with startMode=Manual the performer element contains a defined
   * participant processDefinitionUUID.
   * For an activity with startMode=Automatic the performer element contains the
   * generic participant SYSTEM.
   */
  Set<String> getPerformers();

  /**
   * Deadlines are used to execute a specified java class upon the expiration of
   * a specified period of time.
   * returns (unordered) set of Deadlines.
   * returns empty set if no deadline are defined within the activity.
   */
  Set<DeadlineDefinition> getDeadlines();

  /**
   * Bonita engine allows to specify datafields for activities.
   * As xpdl v1 does not allow this definition, extended attributes are specified
   * for this purpose.
   * @return The (unordered) set of DataFieldDefinition interfaces.
   */
  Set<DataFieldDefinition> getDataFields();
  /**
   * Hooks are user defined logic that can be triggered at some points of the life of
   * the activity.<br>
   * Those points are:
   * <ul>
   * <li>taskOnReady</li>
   * <li>taskOnStart</li>
   * <li>taskOnFinish</li>
   * </ul>
   * <br>
   * If the activity has startMode=Manual then these points of the life of the activity
   * is synchronized with ones of the task.
   * @return The (unordered) set of HookDefinition interfaces.
   */
  List<HookDefinition> getHooks();
  /**
   * Performer assignment allows to perform various assignment rules within the task module.<br>
   * All the users defined into the role type can see and execute this one.
   * By adding this functionality, we can:
   * <br>
   * <ul>
   * <li>assign the task to a user of the role (group) by calling a java class in charge to the user selection
   * into the user group (callback performer assignment)</li>
   * <li>assign dynamically the activity to a user by using an activity variable</li>
   * (property performer assignment)</li>
   * </ul>
   * @return The PerformerAssignDefinition interface.
   */
  FilterDefinition getFilter();

  MultiInstantiationDefinition getMultiInstantiationDefinition();
  
  /**
   * Returns Join Type definition.
   */
  JoinType getJoinType();
  /**
   * Returns Split Type definition.
   */
  SplitType getSplitType();
  
  boolean isAsynchronous();
  
  boolean isSubflow();
  
  Map<String, Collection<String>> getSubflowInParameters();
  Map<String, Collection<String>> getSubflowOutParameters();
  
}
