/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.facade.def.majorElement.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.FilterDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.MultiInstantiationDefinition;
import org.ow2.bonita.facade.def.element.impl.ConnectorDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;

public class ActivityDefinitionImpl extends ProcessElementImpl implements ActivityDefinition {

  private static final long serialVersionUID = 1545928041850807545L;

  protected ActivityDefinitionUUID uuid;
  protected Set<DeadlineDefinition> deadlines;
  protected Set<String> performers;
  protected JoinType joinType;
  protected SplitType splitType;
  protected List<HookDefinition> hooks;
  protected FilterDefinition filter;
  protected Set<DataFieldDefinition> dataFields;

  protected String subflowInParameters;
  protected String subflowOutParameters;
  protected String subflowProcessName;

  protected MultiInstantiationDefinition activityInstantiator;
  protected boolean asynchronous;

  protected ActivityDefinitionImpl() { }

  public ActivityDefinitionImpl(final ProcessDefinitionUUID processUUID,
      final String name, final Set<String> performers,
      final String subflowProcessName) {
    super(name, processUUID);
    this.uuid = new ActivityDefinitionUUID(processUUID, name);
    this.performers = performers;
    this.subflowProcessName = subflowProcessName;
    this.joinType = JoinType.AND;
    this.splitType = SplitType.AND;
    this.dataFields = new HashSet<DataFieldDefinition>();
    this.asynchronous = false;
  }

  public ActivityDefinitionImpl(ActivityDefinition src) {
    super(src);
    this.uuid = new ActivityDefinitionUUID(src.getUUID());
    Set<DeadlineDefinition> deadlines = src.getDeadlines();
    if (deadlines != null) {
      this.deadlines = new HashSet<DeadlineDefinition>();
      for (DeadlineDefinition d : deadlines) {
        this.deadlines.add(new ConnectorDefinitionImpl(d));
      }
    }
    Set<String> performers = src.getPerformers();
    if (performers != null) {
      this.performers = new HashSet<String>();
      for (String performer : performers) {
        this.performers.add(performer);
      }
    }
    this.joinType = src.getJoinType();
    this.splitType = src.getSplitType();
    List<HookDefinition> hooks = src.getHooks();
    if (hooks != null) {
      this.hooks = new ArrayList<HookDefinition>();
      for (HookDefinition d : hooks) {
        this.hooks.add(new ConnectorDefinitionImpl(d));
      }
    }
    if (src.getFilter() != null) {
      this.filter = new ConnectorDefinitionImpl(src.getFilter());
    }
    Set<DataFieldDefinition> dataFields = src.getDataFields();
    if (dataFields != null) {
      this.dataFields = new HashSet<DataFieldDefinition>();
      for (DataFieldDefinition d : dataFields) {
        this.dataFields.add(new DataFieldDefinitionImpl(d));
      }
    }
    if (src.getMultiInstantiationDefinition() != null) {
      this.activityInstantiator = new ConnectorDefinitionImpl(src.getMultiInstantiationDefinition());
    }
    this.asynchronous = src.isAsynchronous();

    ActivityDefinitionImpl srcImpl = (ActivityDefinitionImpl) src;

    this.subflowInParameters = srcImpl.subflowInParameters;
    this.subflowOutParameters = srcImpl.subflowOutParameters;
    this.subflowProcessName = src.getSubflowProcessName();
  }

  public Set<DeadlineDefinition> getDeadlines() {
    return deadlines;
  }

  public Set<String> getPerformers() {
    return performers;
  }

  public String getSubflowProcessName() {
    return subflowProcessName;
  }

  public List<HookDefinition> getHooks() {
    return hooks;
  }

  public FilterDefinition getFilter() {
    return filter;
  }

  public Set<DataFieldDefinition> getDataFields() {
    return dataFields;
  }

  public ActivityDefinitionUUID getUUID() {
    return uuid;
  }

  public MultiInstantiationDefinition getMultiInstantiationDefinition() {
    return activityInstantiator;
  }

  public boolean isAsynchronous() {
    return this.asynchronous;
  }

  public JoinType getJoinType() {
    return this.joinType;
  }

  public SplitType getSplitType() {
    return this.splitType;
  }

  public boolean isAutomatic() {
    return (performers == null || performers.isEmpty())  && filter == null;
  }

  public void addData(DataFieldDefinition data) {
    if (dataFields == null) {
      dataFields = new HashSet<DataFieldDefinition>();
    }
    dataFields.add(data); 
  }

  public void setFilter(FilterDefinition filter) {
    this.filter = filter;
  }

  public void setMultiInstanciation(MultiInstantiationDefinition multiInstanciation) {
    this.activityInstantiator = multiInstanciation;
  }

  public void addDeadline(DeadlineDefinition deadline) {
    if (deadlines == null) {
      deadlines = new HashSet<DeadlineDefinition>();
    }
    this.deadlines.add(deadline);
  }

  public void addConnector(HookDefinition connector) {
    if (hooks == null) {
      hooks = new ArrayList<HookDefinition>();
    }
    this.hooks.add(connector);
  }

  public void setJoinType(JoinType join) {
    this.joinType = join;
  }

  public void setSplitType(SplitType split) {
    this.splitType = split;
  }

  public void setAsynchronous(boolean asynchronous) {
    this.asynchronous = asynchronous;
  }

  private Map<String, Collection<String>> parseParameters(String s) {
    if (s == null) {
      return null;
    }
    String[] couples = s.split(";");
    Map<String, Collection<String>> result = new HashMap<String, Collection<String>>();
    for (String couple : couples) {
      String[] st = couple.split(":");
      String key = st[0];
      String[] values = st[1].split(",");
      Collection<String> c = new HashSet<String>();
      for (String str : values) {
        c.add(str);
      }
      result.put(key, c);
    }
    return result;
  }

  private String convertParameters(Map<String, Collection<String>> map) {
    if (map == null || map.isEmpty()) {
      return null;
    }
    String s = "";
    for (Map.Entry<String, Collection<String>> entry : map.entrySet()) {
      s += entry.getKey();
      s += ":";
      for (String st : entry.getValue()) {
        s += st;
        s += ",";
      }
      s = s.substring(0, s.length() - 1);
      s += ";";
    }
    s = s.substring(0, s.length() - 1);
    return s;
  }
  
  public Map<String, Collection<String>> getSubflowInParameters() {
    if (this.subflowInParameters == null) {
      return Collections.emptyMap();
    }
    return parseParameters(this.subflowInParameters);
  }

  public Map<String, Collection<String>> getSubflowOutParameters() {
    if (this.subflowOutParameters == null) {
      return Collections.emptyMap();
    }
    return parseParameters(this.subflowOutParameters);
  }

  public boolean isSubflow() {
    return getSubflowProcessName() != null;
  }

  public String addParameter(String parameters, String key, String value) {
    Map<String, Collection<String>> map = parseParameters(parameters);
    if (map == null) {
      map = new HashMap<String, Collection<String>>();
    }
    Collection<String> existingValues = map.get(key);
    if (existingValues == null) {
      Collection<String> c = new HashSet<String>();
      c.add(value);
      map.put(key, c);
    } else {
      existingValues.add(value);
    }
    return convertParameters(map);
  }

  public void addSubflowOutParameter(String subProcessDatafieldName, String parentProcessDatafieldName) {
    this.subflowOutParameters = addParameter(this.subflowOutParameters, subProcessDatafieldName, parentProcessDatafieldName);
  }

  public void addSubflowInParameter(String parentProcessDatafieldName, String subProcessDatafieldName) {
    this.subflowInParameters = addParameter(this.subflowInParameters, parentProcessDatafieldName, subProcessDatafieldName);
  }

}
