/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.facade.def.majorElement.impl;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.ow2.bonita.facade.def.element.BusinessArchive;
import org.ow2.bonita.facade.def.element.impl.BusinessArchiveImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.def.majorElement.TransitionDefinition;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.util.CopyTool;
import org.ow2.bonita.util.Misc;

public class ProcessDefinitionImpl extends NamedElementImpl implements ProcessFullDefinition {

  /**
   * 
   */
  private static final long serialVersionUID = -572795239631090498L;

  protected long dbid;

  protected ProcessDefinitionUUID uuid;
  
  protected String version;
  protected ProcessState state;
  protected Set<String> classDependencies;
  protected BusinessArchive businessArchive;
  protected Set<String> processDependencies;

  protected Date deployedDate;
  protected Date undeployedDate;
  protected String deployedBy;
  protected String undeployedBy;

  protected Set<DataFieldDefinition> dataFields;
  protected Set<ParticipantDefinition> participants;
  protected Set<ActivityDefinition> activities;
  protected Set<TransitionDefinition> transitions;

  protected Map<String, String> metadata;
  
  protected ProcessDefinitionImpl() { }

  public ProcessDefinitionImpl(final String name, final String version) {
    super(name);
    Misc.checkArgsNotNull(name, version);
    this.uuid = new ProcessDefinitionUUID(name, version);
    this.version = version;
    if (this.version == null) {
      this.version = "1.0";
    }
  }

  public ProcessDefinitionImpl(final ProcessDefinition src) {
    super(src);
    this.uuid = new ProcessDefinitionUUID(src.getUUID());
    
    this.state = src.getState();
    this.version = src.getVersion();
    this.classDependencies = CopyTool.copy(src.getClassDependencies());    
    this.processDependencies = CopyTool.copy(src.getProcessDependencies());
    this.businessArchive = new BusinessArchiveImpl(src.getBusinessArchive());

    this.deployedDate = CopyTool.copy(src.getDeployedDate());
    this.undeployedDate = CopyTool.copy(src.getUndeployedDate());
    this.deployedBy = src.getDeployedBy();
    this.undeployedBy = src.getUndeployedBy();

    Set<DataFieldDefinition> dataFields = src.getDataFields();
    if (dataFields != null) {
      this.dataFields = new HashSet<DataFieldDefinition>();
      for (DataFieldDefinition d : dataFields) {
        this.dataFields.add(new DataFieldDefinitionImpl(d));
      }
    }
    Set<ParticipantDefinition> participants = src.getParticipants();
    if (participants != null) {
      this.participants = new HashSet<ParticipantDefinition>();
      for (ParticipantDefinition d : participants) {
        this.participants.add(new ParticipantDefinitionImpl(d));
      }
    }
    Set<ActivityDefinition> activities = src.getActivities();
    if (activities != null) {
      this.activities = new HashSet<ActivityDefinition>();
      for (ActivityDefinition d : activities) {
        this.activities.add(new ActivityDefinitionImpl(d));
      }
    }
    Set<TransitionDefinition> transitions = src.getTransitions();
    if (transitions != null) {
      this.transitions = new HashSet<TransitionDefinition>();
      for (TransitionDefinition d : transitions) {
        this.transitions.add(new TransitionDefinitionImpl(d));
      }
    }

    this.metadata = new HashMap<String, String>();
    final Map<String, String> meta = src.getMetaData();
    if (meta != null) {
      for (Entry<String, String> entry : meta.entrySet()) {
        this.metadata.put(entry.getKey(), entry.getValue());
      }
    }
  }

  @Override
  public String toString() {
    String st = this.getClass().getName()
    + "[uuid: " + getUUID()
    + ", name:" + getName()
    + ", description:" + getDescription()
    + ", version:" + getVersion();
    st +=  "]";
    return st;
  }

  public ProcessState getState() {
    return this.state;
  }

  public String getVersion() {
    return this.version;
  }

  public ProcessDefinitionUUID getUUID() {
    return this.uuid;
  }

  public BusinessArchive getBusinessArchive() {
    return this.businessArchive;
  }
  public Set<String> getClassDependencies() {
    return this.classDependencies;
  }

  public Set<String> getProcessDependencies() {
    return this.processDependencies;
  }

  public Date getDeployedDate() {
    return this.deployedDate;
  }

  public String getDeployedBy() {
    return this.deployedBy;
  }

  public Date getUndeployedDate() {
    return this.undeployedDate;
  }


  public String getUndeployedBy() {
    return this.undeployedBy;
  }

  public Set<DataFieldDefinition> getDataFields() {
    if (dataFields == null) {
      return Collections.emptySet();
    }
    return dataFields;
  }

  public Set<ParticipantDefinition> getParticipants() {
    if (participants == null) {
      return Collections.emptySet();
    }
    return participants;
  }

  public Set<ActivityDefinition> getActivities() {
    if (activities == null) {
      return Collections.emptySet();
    }
    return activities;
  }

  public Set<TransitionDefinition> getTransitions() {
    if (transitions == null) {
      return Collections.emptySet();
    }
    return transitions;
  }

  public Map<String, String> getMetaData() {
    return metadata;
  }
  public String getAMetaData(String key) {
    Misc.checkArgsNotNull(key);
    if (this.metadata == null) {
      return null;
    }
    return metadata.get(key);
  }


  /**
   * SETTERS
   */
  public void addProcessDependency(final String subProcessName) {
    if (processDependencies == null) {
      processDependencies = new HashSet<String>();
    }
    processDependencies.add(subProcessName);
  }

  public void addClassDependency(final String className) {
    if (classDependencies == null) {
      classDependencies = new HashSet<String>();
    }
    classDependencies.add(className);
  }

  public void setState(ProcessState state) {
    this.state = state; 
  }

  public void setUndeployedDate(Date undeployedDate) {
    this.undeployedDate = undeployedDate;
  }
  public void setUndeployedBy(String undeployedBy) {
    this.undeployedBy = undeployedBy;
  }

  public void addData(DataFieldDefinition data) {
    if (dataFields == null) {
      dataFields = new HashSet<DataFieldDefinition>();
    }
    this.dataFields.add(data);
  }

  public void addGroup(ParticipantDefinitionImpl group) {
    if (participants == null) {
      participants = new HashSet<ParticipantDefinition>();
    }
    this.participants.add(group);
  }

  public void addTranistion(TransitionDefinition transition) {
    if (transitions == null) {
      transitions = new HashSet<TransitionDefinition>();
    }
    this.transitions.add(transition);
  }

  public void addActivity(ActivityDefinition activity) {
    if (activities == null) {
      activities = new HashSet<ActivityDefinition>();
    }
    this.activities.add(activity);
  }

  public void setDeployedDate(Date deployedDate) {
    this.deployedDate = deployedDate;
  }

  public void setDeployedBy(String deployedBy) {
    this.deployedBy = deployedBy;
  }

  public void setBusinessArchive(BusinessArchive businessArchive) {
    this.businessArchive = businessArchive;
  }

  public void deleteAMetaData(String key) {
    Misc.checkArgsNotNull(key);
    metadata.remove(key);
  }

  public void addAMetaData(String key, String value) {
    Misc.checkArgsNotNull(key, value);
    if (this.metadata == null) {
      this.metadata = new HashMap<String, String>();
    }
    this.metadata.put(key, value);
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj.getClass().equals(this.getClass()))) {
      return false;
    }
    final ProcessDefinitionImpl other = (ProcessDefinitionImpl)obj;
    return other.getUUID().equals(this.uuid);
  }

  public ActivityDefinition getActivity(final String name) {
    for (ActivityDefinition activity : getActivities()) {
      if (activity.getName().equals(name)) {
        return activity;
      }
    }
    return null;
  }
  public Collection<ActivityDefinition> getFirstActivities() {
    Set<String> notFirstActivities = new HashSet<String>();

    for (TransitionDefinition transition : getTransitions()) {
      if (transition.getTo() != null) {
        notFirstActivities.add(transition.getTo());
      }
    }
    Set<ActivityDefinition> firstActivities = new HashSet<ActivityDefinition>();
    for (ActivityDefinition activity : getActivities()) {
      if (!notFirstActivities.contains(activity.getName())) {
        firstActivities.add(activity);
      }
    }
    return firstActivities;
  }

  @Override
  public int hashCode() {
    return this.uuid.hashCode();
  }
}
