/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Charles Souillard, BonitaSoft
 **/
package org.ow2.bonita.facade.runtime.impl;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.runtime.InstanceStateUpdate;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.VariableUpdate;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;

/**
 * @author Pierre Vigneras, Charles Souillard
 */
public class ProcessInstanceImpl extends RuntimeRecordImpl implements ProcessInstance {

  private static final long serialVersionUID = 8366284714927360659L;

  protected ProcessInstanceUUID parentInstanceUUID;
  protected Set<ProcessInstanceUUID> childrenInstanceUUID;
  protected Map<String, Object> initialVariableValues;

  protected List<VariableUpdate> variableUpdates = new ArrayList<VariableUpdate>();
  protected List<InstanceStateUpdate> instanceStateUpdates = new ArrayList<InstanceStateUpdate>();

  @Override
  public String toString() {
    return this.getClass().getName()
      + "[uuid: " + getUUID()
      + ", processDefinitionUUID: " + getProcessDefinitionUUID()
      + ", processUUID: " + getProcessInstanceUUID()
      + ", parentInstanceUUID: " + getParentInstanceUUID()
      + ", childrenInstanceUUID: " + getChildrenInstanceUUID()
      + ", startedBy: " + getStartedBy()
      + ", endedBy: " + getEndedBy()
      + ", startedDate: " + getStartedDate()
      + ", endedDate: " + getEndedDate()
      + ", initialVariableValues: " + getInitialVariableValues()
      + "]";
  }

  protected ProcessInstanceImpl() {
    super();
  }

  public ProcessInstanceImpl(final ProcessInstanceUUID instanceUUID,
      final ProcessDefinitionUUID processUUID,
      final ProcessInstanceUUID parentInstanceUUID,
      final Map<String, Object> initialVariableValues) {
    super(processUUID, instanceUUID);
    this.parentInstanceUUID = parentInstanceUUID;
    this.initialVariableValues = initialVariableValues;
  }

  public ProcessInstanceImpl(final ProcessInstance processInstance) {
    super(processInstance);
    this.parentInstanceUUID = processInstance.getParentInstanceUUID();

    final Set<ProcessInstanceUUID> children = processInstance.getChildrenInstanceUUID();
    if (children != null && !children.isEmpty()) {
      this.childrenInstanceUUID = new HashSet<ProcessInstanceUUID>();
      for (final ProcessInstanceUUID child : children) {
        this.childrenInstanceUUID.add(child);
      }
    }

    this.initialVariableValues = processInstance.getInitialVariableValues();

    final List<VariableUpdate> list = processInstance.getVariableUpdates();
    if (list != null && !list.isEmpty()) {
      this.variableUpdates = new ArrayList<VariableUpdate>();
      for (final VariableUpdate varUpdate : list) {
        this.variableUpdates.add(new VariableUpdateImpl(varUpdate));
      }
    }
    final List<InstanceStateUpdate> lisu = processInstance.getInstanceStateUpdates();
    if (lisu != null && !lisu.isEmpty()) {
      this.instanceStateUpdates = new ArrayList<InstanceStateUpdate>();
      for (final InstanceStateUpdate isu : lisu) {
        this.instanceStateUpdates.add(new InstanceStateUpdateImpl(isu));
      }
    }
  }

  private InstanceStateUpdate getFirstStateUpdate(final InstanceState... instanceStates) {
    if (this.instanceStateUpdates != null && instanceStates != null && instanceStates.length > 0) {
      for (InstanceStateUpdate su : this.instanceStateUpdates) {
        InstanceState state = su.getInstanceState();
        for (InstanceState is : instanceStates) {
          if (is.equals(state)) {
            return su;
          }
        }
      }
    }
    return null;
  }
  
  public String getEndedBy() {
    InstanceStateUpdate is = getFirstStateUpdate(InstanceState.FINISHED, InstanceState.ABORTED, InstanceState.CANCELLED);
    if (is != null) {
      return is.getUpdatedBy();
    }
    return null;
  }

  public Date getEndedDate() {
    InstanceStateUpdate is = getFirstStateUpdate(InstanceState.FINISHED, InstanceState.ABORTED, InstanceState.CANCELLED);
    if (is != null) {
      return is.getUpdatedDate();
    }
    return null;
  }


  public String getStartedBy() {
    InstanceStateUpdate is = getFirstStateUpdate(InstanceState.STARTED);
    if (is != null) {
      return is.getUpdatedBy();
    }
    return null;
  }

  public Date getStartedDate() {
    InstanceStateUpdate is = getFirstStateUpdate(InstanceState.STARTED);
    if (is != null) {
      return is.getUpdatedDate();
    }
    return null;
  }
  
  public Map<String, Object> getInitialVariableValues() {
    return this.initialVariableValues;
  }

  public InstanceState getInstanceState() {
    if (getInstanceStateUpdates() != null && !getInstanceStateUpdates().isEmpty()) {
      final InstanceStateUpdate stateUpdate = getInstanceStateUpdates().get(getInstanceStateUpdates().size() - 1);
      if (stateUpdate.getInstanceState() != null) {
        return stateUpdate.getInstanceState();
      }
    }
    return InstanceState.STARTED;
  }

  public ProcessInstanceUUID getParentInstanceUUID() {
    return this.parentInstanceUUID;
  }

  public Set<ProcessInstanceUUID> getChildrenInstanceUUID() {
    if (this.childrenInstanceUUID == null) {
      return new HashSet<ProcessInstanceUUID>();
    }
    return this.childrenInstanceUUID;
  }

  public Object getInitialVariableValue(final String variableId) {
    return getInitialVariableValues().get(variableId);
  }

  public Map<String, Object> getLastKnownVariableValues() {
    Map<String, Object> var = getInitialVariableValues();
    if (var != null) {
      var = new HashMap<String, Object>(var);
    } else {
      var = new HashMap<String, Object>();
    }
    for (final VariableUpdate varUp : getVariableUpdates()) {
      var.put(varUp.getName(), varUp.getValue());
    }
    return var;
  }


  public ProcessInstanceUUID getUUID() {
    return getProcessInstanceUUID();
  }

  public List<VariableUpdate> getVariableUpdates() {
    return this.variableUpdates;
  }

  public List<InstanceStateUpdate> getInstanceStateUpdates() {
    return this.instanceStateUpdates;
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj.getClass().equals(this.getClass()))) {
      return false;
    }
    final ProcessInstanceImpl other = (ProcessInstanceImpl)obj;
    if (other.getUUID() == null) {
      return getUUID() == null;
    }
    return other.getUUID().equals(getUUID());
  }

  @Override
  public int hashCode() {
    return getUUID().hashCode();
  }
}
