/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.identity.auth;

import java.security.Principal;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.security.auth.Destroyable;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;

/**
 * @author "Pierre Vigneras"
 */
public class PlainLoginModule implements LoginModule {

  private static final String NAME_PROMPT = "Name: ";
  private static final String PASSWORD_PROMPT = "Password: ";

  private static final String JAVAX_SECURITY_AUTH_LOGIN_PASSWORD = "javax.security.auth.login.password";
  private static final String JAVAX_SECURITY_AUTH_LOGIN_NAME = "javax.security.auth.login.name";

  /**
   * Property key for the debug flag. Defined to be "debug".
   * 
   * Property Value. If set, should be either "true" or "false". Default is
   * "false".
   */
  public static final String DEBUG_OPTION_NAME = "debug";
  private static final Object LOGINS_OPTION_NAME = "logins";
  private static final Object PASSWORDS_OPTION_NAME = "passwords";
  private Subject subject = null;
  private CallbackHandler callbackHandler = null;
  private Map<String, Object> sharedState;
  private Map<String, Object> options = null;
  private boolean debug = false;

  private String id;
  private List<String> logins;
  private List<String> passwords;

  /**
   * Initialize this LoginModule. This method is called by the LoginContext
   * after this LoginModule has been instantiated. The purpose of this method is
   * to initialize this LoginModule with the relevant information. If this
   * LoginModule does not understand any of the data stored in sharedState or
   * options parameters, they can be ignored.
   * 
   * @param subject
   *          the Subject to be authenticated.
   * @param callbackHandler
   *          a CallbackHandler for communicating with the end user (prompting
   *          for usernames and passwords, for example).
   * @param sharedState
   *          state shared with other configured LoginModules.
   * @param options
   *          options specified in the login Configuration for this particular
   *          LoginModule.
   */
  @SuppressWarnings("unchecked")
  public void initialize(final Subject subject,
      final CallbackHandler callbackHandler, final Map<String, ? > sharedState,
      final Map<String, ? > options) {

    Misc.checkArgsNotNull(subject, callbackHandler, sharedState, options);
    
    this.subject = subject;
    this.callbackHandler = callbackHandler;
    this.sharedState = (Map<String, Object>) sharedState;
    this.options = (Map<String, Object>) options;
    final String debugFlag = (String) options
        .get(PlainLoginModule.DEBUG_OPTION_NAME);
    if (debugFlag != null) {
      this.debug = Boolean.valueOf(debugFlag);
    }
  }

  /**
   * Method to authenticate a Subject (phase 1). The implementation of this
   * method authenticates a Subject. For example, it may prompt for Subject
   * information such as a username and password and then attempt to verify the
   * password. This method saves the result of the authentication attempt as
   * private state within the LoginModule.
   * 
   * @return true if the authentication succeeded, or false if this LoginModule
   *         should be ignored.
   * @throws LoginException
   *           if the authentication fails
   */
  public boolean login() throws LoginException {
    if (this.debug) {
      System.err.println("[" + PlainLoginModule.class.getName()
          + "] login() - preparing - step 1");
    }
    if (this.logins == null) {
      final String loginList = (String) this.options
          .get(PlainLoginModule.LOGINS_OPTION_NAME);
      final String passwordList = (String) this.options
          .get(PlainLoginModule.PASSWORDS_OPTION_NAME);
      if (loginList == null) {
      	String message = ExceptionManager.getInstance().getFullMessage(
      			"bi_PLM_1", PlainLoginModule.LOGINS_OPTION_NAME);
        throw new MissingOptionException("bi_PLM_1", message);
      }
      if (passwordList == null) {
      	String message = ExceptionManager.getInstance().getFullMessage(
      			"bi_PLM_2", PlainLoginModule.PASSWORDS_OPTION_NAME);
        throw new MissingOptionException("bi_PLM_2",message);
      }
      final StringTokenizer loginTokenizer = new StringTokenizer(loginList, ",");
      final StringTokenizer passwordTokenizer = new StringTokenizer(
          passwordList, ",");
      if (loginTokenizer.countTokens() != passwordTokenizer.countTokens()) {
      	String message = ExceptionManager.getInstance().getFullMessage("bi_PLM_3");
      	throw new BadLoginOptionException(message);
      }
      this.logins = new ArrayList<String>();
      this.passwords = new ArrayList<String>();
      while (loginTokenizer.hasMoreTokens()) {
        final String name = loginTokenizer.nextToken().trim();
        if (this.logins.contains(name)) {
        	String message = ExceptionManager.getInstance().getFullMessage(
        			"bi_PLM_4", name);
          throw new BadLoginOptionException(message);
        }
        final String password = passwordTokenizer.nextToken().trim();
        this.logins.add(name);
        this.passwords.add(password);
      }
    }
    try {
      String name = (String) this.sharedState
          .get(PlainLoginModule.JAVAX_SECURITY_AUTH_LOGIN_NAME);
      String password = (String) this.sharedState
          .get(PlainLoginModule.JAVAX_SECURITY_AUTH_LOGIN_PASSWORD);
      final List<Callback> callbacks = new ArrayList<Callback>();
      final NameCallback nameCallback = new NameCallback(
          PlainLoginModule.NAME_PROMPT);
      final PasswordCallback passwordCallback = new PasswordCallback(
          PlainLoginModule.PASSWORD_PROMPT, false);
      if (name == null) {
        callbacks.add(nameCallback);
      }
      if (password == null) {
        callbacks.add(passwordCallback);
      }
      if (!callbacks.isEmpty()) {
        if (this.debug) {
          System.err.println("[" + PlainLoginModule.class.getName()
              + "] login() - callback - step 2");
        }
        this.callbackHandler.handle(callbacks.toArray(new Callback[0]));
        if (name == null) {
          name = nameCallback.getName();
        }
        if (password == null) {
          password = new String(passwordCallback.getPassword());
          passwordCallback.clearPassword();
        }
      }
      if (this.debug) {
        System.err.println("[" + PlainLoginModule.class.getName()
            + "] login() - authenticating - step 3");
      }
      final int i = this.logins.indexOf(name);
      if (i != -1) {
        if (this.passwords.get(i).equals(password)) {
          this.id = name;
        }
      }
      // TODO: what is Step 4 then?!
      if (this.debug) {
        System.err.println("[" + PlainLoginModule.class.getName()
            + "] login() - storing data - step 5");
      }
      this.sharedState.put(PlainLoginModule.JAVAX_SECURITY_AUTH_LOGIN_NAME,
          name);
      this.sharedState.put(PlainLoginModule.JAVAX_SECURITY_AUTH_LOGIN_PASSWORD,
          password);
      if (this.debug) {
        System.err.println("[" + PlainLoginModule.class.getName()
            + "] login() - returning - step 6");
      }
      if (this.id == null) {
      	String message = ExceptionManager.getInstance().getFullMessage("bi_PLM_5");
        throw new FailedLoginException(message);
      }
      return true;
    } catch (final Exception e) {
      final LoginException le = new LoginException();
      le.initCause(e);
      throw le;
    }
  }

  /**
   * Method to commit the authentication process (phase 2). This method is
   * called if the LoginContext's overall authentication succeeded (the relevant
   * REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules succeeded). If
   * this LoginModule's own authentication attempt succeeded (checked by
   * retrieving the private state saved by the login method), then this method
   * associates relevant Principals and Credentials with the Subject located in
   * the LoginModule. If this LoginModule's own authentication attempted failed,
   * then this method removes/destroys any state that was originally saved.
   * 
   * @return true if this method succeeded, or false if this LoginModule should
   *         be ignored.
   * @throws LoginException
   *           if the commit fails
   */
  public boolean commit() throws LoginException {
    if (this.id == null) {
    	String message = ExceptionManager.getInstance().getFullMessage("bi_PLM_6");
      throw new FailedLoginException(message);
    }
    final Set<Principal> principals = this.subject.getPrincipals();
    principals.add(new BonitaPrincipal(this.id));
    return true;
  }

  /**
   * Method to abort the authentication process (phase 2). This method is called
   * if the LoginContext's overall authentication failed. (the relevant
   * REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules did not succeed).
   * If this LoginModule's own authentication attempt succeeded (checked by
   * retrieving the private state saved by the login method), then this method
   * cleans up any state that was originally saved.
   * 
   * @return true if this method succeeded, or false if this LoginModule should
   *         be ignored.
   * @throws LoginException
   *           if the abort fails
   */
  public boolean abort() throws LoginException {
    if (this.debug) {
      System.err.println("[" + PlainLoginModule.class.getName() + "] abort()");
    }
    if (this.id == null) {
      return false;
    }
    this.subject = null;
    this.id = null;
    return true;
  }

  /**
   * Method which logs out a Subject. An implementation of this method might
   * remove/destroy a Subject's Principals and Credentials.
   * 
   * @return true if this method succeeded, or false if this LoginModule should
   *         be ignored.
   * @throws LoginException
   *           if the logout fails
   */
  public boolean logout() throws LoginException {
    if (this.id != null) {
      if (this.debug) {
        System.err.println("[" + PlainLoginModule.class.getName()
            + "] logout() - removing principals");
      }
      // Remove only principals added by our commit method
      final Set<Principal> principals = new HashSet<Principal>(this.subject
          .getPrincipals());
      for (final Principal p : principals) {
        if (p instanceof BonitaPrincipal) {
          if (this.debug) {
            System.err.println("[" + PlainLoginModule.class.getName()
                + "] logout() - removing principal: " + p);
          }
          this.subject.getPrincipals().remove(p);
        }
      }
      if (this.debug) {
        System.err.println("[" + PlainLoginModule.class.getName()
            + "] logout() - destroying/removing credentials");
      }
      // Remove/destroy only credentials added by our commit method
      final Set<Object> credentials = new HashSet<Object>(this.subject
          .getPublicCredentials());
      for (final Object o : credentials) {
        if (o instanceof Destroyable) {
          if (this.debug) {
            System.err.println("[" + PlainLoginModule.class.getName()
                + "] logout() - destroying credential: " + o);
          }
          // Bug: only from this module !!
          // ((Destroyable) o).destroy();
        }
        if (!this.subject.isReadOnly()) {
          if (this.debug) {
            System.err.println("[" + PlainLoginModule.class.getName()
                + "] logout() - removing credential: " + o);
          }
          this.subject.getPublicCredentials().remove(o);
        }
      }
    }
    return true;
  }
}
