/**
 * Copyright (C) 2009  BonitaSoft S.A..
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.io.Serializable;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import org.ow2.bonita.facade.def.element.FilterDefinition;
import org.ow2.bonita.facade.def.element.RoleMapperDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition.Event;
import org.ow2.bonita.facade.def.element.impl.ConnectorDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.TransitionDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition.JoinType;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition.SplitType;
import org.ow2.bonita.facade.def.majorElement.impl.ActivityDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.DataFieldDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.DescriptionElementImpl;
import org.ow2.bonita.facade.def.majorElement.impl.NamedElementImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ParticipantDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.TransitionDefinitionImpl;
import org.ow2.bonita.parsing.XpdlParser;
import org.ow2.bonita.util.xml.Parse;
import org.ow2.bonita.util.xml.Problem;

/**
 * 
 * @author Matthieu Chaffotte, Charles Souillard
 *
 */
public class ProcessBuilder {

  private static Collection<Problem> problems;
  private static Stack<Object> stack;

  /**
   * Creates a process with a unique name and a process version
   * @param name the unique name of the process
   * @return ProcessBuilder in order to finish the process build
   */
  public static ProcessBuilder createProcess(String name, String version) {
    problems = new ArrayList<Problem>();
    stack = new Stack<Object>();
    return new ProcessBuilder(name, version);
  }

  public static ProcessDefinition createProcessFromXpdlFile(final URL xpdlUrl) {
    final Parse parse = new XpdlParser().createParse();
    parse.setUrl(xpdlUrl);

    final ProcessDefinition process = (ProcessDefinition) parse.execute().getDocumentObject();

    Misc.showProblems(parse.getProblems(), "xpdl file");

    //parse.popObject();
    Misc.badStateIfNull(process, "Ouch! The returned Clientprocess is null!");
    return process;
  }

  private ProcessBuilder(String processame, String processVersion) {
    Misc.checkArgsNotNull(processame);
    if ("".equals(processame.trim())) {
      problems.add(new Problem("Process name is an empty string.", Problem.SEVERITY_ERROR));
    } if (processVersion == null || "".equals(processVersion.trim())) {
      processVersion = "1.0";
    }
    ProcessDefinition process = new ProcessDefinitionImpl(processame, processVersion);
    push(process);
  }

  /**
   * Adds a description to the current BPM entity
   * @param description the BPM description
   * @return ProcessBuilder in order to finish the process build
   */
  public ProcessBuilder addDescription(String description) {
    Object obj = peek(DescriptionElementImpl.class);
    if (isDescriptionElement(obj)) {
      ((DescriptionElementImpl) obj).setDescription(description);
    } else {
      problems.add(new Problem("Unable to set description " + description + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a label to the current BPM entity
   * @param label the BPM label
   * @return ProcessBuilder in order to finish the process build
   */
  public ProcessBuilder addLabel(final String label) {
    Object obj = peek(NamedElementImpl.class);
    if (isNamedElement(obj)) {
      ((NamedElementImpl) obj).setLabel(label);
    } else {
      problems.add(new Problem("Unable to set label " + label + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }
  
  public ProcessBuilder addCharData(String dataName) {
    return addCharData(dataName, null);
  }
  
  public ProcessBuilder addShortData(String dataName) {
    return addShortData(dataName, null);
  }
  
  public ProcessBuilder addLongData(String dataName) {
    return addLongData(dataName, null);
  }
  
  public ProcessBuilder addDoubleData(String dataName) {
    return addDoubleData(dataName, null);
  }
  
  public ProcessBuilder addFloatData(String dataName) {
    return addFloatData(dataName, null);
  }
  
  public ProcessBuilder addBooleanData(String dataName) {
    return addBooleanData(dataName, null);
  }
  
  public ProcessBuilder addIntegerData(String dataName) {
    return addIntegerData(dataName, null);
  }
  
  public ProcessBuilder addStringData(String dataName) {
    return addStringData(dataName, null);
  }
  
  public ProcessBuilder addDateData(String dataName) {
    return addDateData(dataName, null);
  }
  
  public ProcessBuilder addCharData(String dataName, Character initialValue) {
    return addObjectData(dataName, Character.class.getName(), initialValue);
  }
  
  public ProcessBuilder addShortData(String dataName, Short initialValue) {
    return addObjectData(dataName, Short.class.getName(), initialValue);
  }
  
  public ProcessBuilder addLongData(String dataName, Long initialValue) {
    return addObjectData(dataName, Long.class.getName(), initialValue);
  }
  
  public ProcessBuilder addDoubleData(String dataName, Double initialValue) {
    return addObjectData(dataName, Double.class.getName(), initialValue);
  }
  
  public ProcessBuilder addFloatData(String dataName, Float initialValue) {
    return addObjectData(dataName, Float.class.getName(), initialValue);
  }
  
  public ProcessBuilder addBooleanData(String dataName, Boolean initialValue) {
    return addObjectData(dataName, Boolean.class.getName(), initialValue);
  }
  
  public ProcessBuilder addIntegerData(String dataName, Integer initialValue) {
    return addObjectData(dataName, Integer.class.getName(), initialValue);
  }
  
  public ProcessBuilder addStringData(String dataName, String initialValue) {
    return addObjectData(dataName, String.class.getName(), initialValue);
  }
  
  public ProcessBuilder addDateData(String dataName, Date initialValue) {
    return addObjectData(dataName, Date.class.getName(), initialValue);
  }

  public ProcessBuilder addEnumData(String dataName, Set<String> enumeariontValues, String initialValue) {
    return addObjectData(dataName, String.class.getName(), initialValue, enumeariontValues);
  }

  private ProcessBuilder addObjectData(String dataName, String dataTypeClassName, Serializable initialValue, Set<String> enumerationValues) {
    Misc.checkArgsNotNull(dataName, dataTypeClassName);
    if ("".equals(dataName.trim())) {
      problems.add(new Problem("Data name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(dataTypeClassName.trim())) {
      problems.add(new Problem("Data type class name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ActivityDefinitionImpl.class);
    DataFieldDefinitionImpl data = null;
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      data = new DataFieldDefinitionImpl(activity.getUUID(), dataName, dataTypeClassName);
      activity.addData(data);
      push(data);
    } else if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      data = new DataFieldDefinitionImpl(process.getUUID(), dataName, dataTypeClassName);
      process.addData(data);
      push(data);
    }
    data.setInitialValue(initialValue);
    data.setEnumerationValues(enumerationValues);
    return this;
  }
  
  private ProcessBuilder addObjectData(String dataName, String dataTypeClassName, Serializable initialValue) {
    return addObjectData(dataName, dataTypeClassName, initialValue, null);
  }

  /**
   * Adds a simple data to a process or an activity.
   * If the last main Element is not an activity
   * the data will be a process one.
   * @param dataName the data name
   * @param type the data type
   * @return
   */
  public ProcessBuilder addData(String dataName, String dataTypeClassName, Set<String> enumerationValues) {
    Misc.checkArgsNotNull(dataName, dataTypeClassName);
    if ("".equals(dataName.trim())) {
      problems.add(new Problem("Data name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(dataTypeClassName.trim())) {
      problems.add(new Problem("Data type class name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ActivityDefinitionImpl.class);
    DataFieldDefinitionImpl data = null;
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      data = new DataFieldDefinitionImpl(activity.getUUID(), dataName, dataTypeClassName);
      activity.addData(data);
      push(data);
    } else if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      data = new DataFieldDefinitionImpl(process.getUUID(), dataName, dataTypeClassName);
      process.addData(data);
      push(data);
    } else {
      problems.add(new Problem("Unable to add data " + dataName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    if (enumerationValues != null) {
      data.setEnumerationValues(enumerationValues);
    }
    return this;
  }

  /**
   * Adds an enumeration data to a process or an activity.
   * If the last main Element is not an activity
   * the data will be a process one.
   * @param dataName the data name
   * @param type the data type
   * @return
   */
  public ProcessBuilder addData(String dataName, String dataTypeClassName) {
    return addData(dataName, dataTypeClassName, null);
  }

  /**
   * Adds a group
   * @param groupName
   * @param groupResolverClassName
   * @return
   */
  public ProcessBuilder addGroup(String groupName) {
    Misc.checkArgsNotNull(groupName);
    if ("".equals(groupName.trim())) {
      problems.add(new Problem("Group name is an empty string.", Problem.SEVERITY_ERROR));
    }

    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      ParticipantDefinitionImpl group = new ParticipantDefinitionImpl(process.getUUID(), groupName);
      process.addGroup(group);
      push(group);
    } else {
      problems.add(new Problem("Unable to add group " + groupName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }
  
  /**
   * Adds a group
   * @param groupName
   * @param groupResolverClassName
   * @return
   */
  public ProcessBuilder addGroupResolver(String groupResolverClassName) {
    Misc.checkArgsNotNull(groupResolverClassName);
    if ("".equals(groupResolverClassName.trim())) {
      problems.add(new Problem("Resolver class name is an empty string.", Problem.SEVERITY_ERROR));
    }

    Object obj = peek(ParticipantDefinitionImpl.class);
    if (isParticipant(obj)) {
      ParticipantDefinitionImpl group = (ParticipantDefinitionImpl) obj;
      RoleMapperDefinition resolver = new ConnectorDefinitionImpl(groupResolverClassName);
      group.setResolver(resolver);
      ProcessDefinitionImpl process = getProcess();
      process.addGroup(group);
      process.addClassDependency(groupResolverClassName);
      push(group);
      push(resolver);
    } else {
      problems.add(new Problem("Unable to add resolver " + groupResolverClassName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a human
   * @param humanName
   * @return
   */
  public ProcessBuilder addHuman(String humanName) {
    Misc.checkArgsNotNull(humanName);
    if ("".equals(humanName.trim())) {
      problems.add(new Problem("Human name is an empty string.", Problem.SEVERITY_ERROR));
    }

    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      ParticipantDefinitionImpl group = new ParticipantDefinitionImpl(process.getUUID(), humanName);
      process.addGroup(group);
      push(group);
    } else {
      problems.add(new Problem("Unable to add human " + humanName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }

    return this;
  }

  /**
   * Adds a system task
   * By default The join and the split types are AND
   * @param taskName
   * @return
   */
  public ProcessBuilder addSystemTask(String taskName) {
    Misc.checkArgsNotNull(taskName);
    if ("".equals(taskName.trim())) {
      problems.add(new Problem("Task name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      ActivityDefinitionImpl activity = new ActivityDefinitionImpl(process.getUUID(), taskName, null, null);
      process.addActivity(activity);
      push(activity);
    } else {
      problems.add(new Problem("Unable to add system task " + taskName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    
    return this;
  }

  /**
   * Add a human task
   * By default The join and the split types are AND
   * @param taskName the human task name
   * @param authorityNames the groupName or a user name
   * @return
   */
  public ProcessBuilder addHumanTask(String taskName, String... authorityNames) {
    Misc.checkArgsNotNull(taskName, authorityNames);
    if ("".equals(taskName.trim())) {
      problems.add(new Problem("Task name is an empty string.", Problem.SEVERITY_ERROR));
    }
    for (String authorityName : authorityNames) {
      if ("".equals(authorityName.trim())) {
        problems.add(new Problem("Authority name is an empty string.", Problem.SEVERITY_ERROR));
      }
    }
    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      Set<String> groups = new HashSet<String>();
      for (String authorityName : authorityNames) {
        groups.add(authorityName);
      }
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      ActivityDefinitionImpl activity = new ActivityDefinitionImpl(process.getUUID(), taskName, groups, null);
      process.addActivity(activity);
      push(activity);
    } else {
      problems.add(new Problem("Unable to add human task " + taskName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a subProcess
   * By default The join and the split types are AND
   * @param taskName
   * @param subProcessParameters
   * @param execution
   * @return
   */
  public ProcessBuilder addSubProcess(String taskName, String processName) {
    Misc.checkArgsNotNull(taskName);
    if ("".equals(taskName.trim())) {
      problems.add(new Problem("Task name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(processName.trim())) {
      problems.add(new Problem("Subprocess name is an empty string.", Problem.SEVERITY_ERROR));
    }
    
    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      ActivityDefinitionImpl activity = new ActivityDefinitionImpl(process.getUUID(), taskName, null, processName);
      process.addActivity(activity);
      process.addProcessDependency(processName);
      push(activity);
    } else {
      problems.add(new Problem("Unable to add sub process " + taskName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }
  
  /**
   * Adds a subProcess in parameter
   */
  public ProcessBuilder addSubProcessInParameter(String parentProcessDatafieldName, String subProcessDatafieldName) {
    Misc.checkArgsNotNull(parentProcessDatafieldName, subProcessDatafieldName);
    if ("".equals(parentProcessDatafieldName.trim())) {
      problems.add(new Problem("Parent process datafield name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(subProcessDatafieldName.trim())) {
      problems.add(new Problem("Sub process datafield name is an empty string.", Problem.SEVERITY_ERROR));
    }
    
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      activity.addSubflowInParameter(parentProcessDatafieldName, subProcessDatafieldName);
    } else {
      problems.add(new Problem("Unable to add sub process parameter {" + parentProcessDatafieldName + ", " + subProcessDatafieldName + "} on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }
  
  /**
   * Adds a subProcess out parameter
   */
  public ProcessBuilder addSubProcessOutParameter(String subProcessDatafieldName, String parentProcessDatafieldName) {
    Misc.checkArgsNotNull(parentProcessDatafieldName, subProcessDatafieldName);
    if ("".equals(subProcessDatafieldName.trim())) {
      problems.add(new Problem("Sub process datafield name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(parentProcessDatafieldName.trim())) {
      problems.add(new Problem("Parent process datafield name is an empty string.", Problem.SEVERITY_ERROR));
    }
    
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      activity.addSubflowOutParameter(subProcessDatafieldName, parentProcessDatafieldName);
    } else {
      problems.add(new Problem("Unable to add sub process parameter {" + parentProcessDatafieldName + ", " + subProcessDatafieldName + "} on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a Route
   * By default The join and the split types are AND
   * @param activityName
   * @return
   */
  public ProcessBuilder addDecisionNode(String taskName) {
    Misc.checkArgsNotNull(taskName);
    if ("".equals(taskName.trim())) {
      problems.add(new Problem("Task name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      ActivityDefinitionImpl activity = new ActivityDefinitionImpl(process.getUUID(), taskName, null, null);
      process.addActivity(activity);
      push(activity);
    } else {
      problems.add(new Problem("Unable to add decision node " + taskName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a deadline to the current activity
   * If the current element is not a activity, the deadline is not set
   * @param condition
   * @param connectorClassName
   * @return
   */
  public ProcessBuilder addDeadline(String condition, String connectorClassName) {
    Misc.checkArgsNotNull(condition, connectorClassName);
    if ("".equals(condition.trim())) {
      problems.add(new Problem("Condition is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(connectorClassName.trim())) {
      problems.add(new Problem("Connector class name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      ConnectorDefinitionImpl deadline = new ConnectorDefinitionImpl(connectorClassName);
      deadline.setCondition(condition);
      activity.addDeadline(deadline);
      ProcessDefinitionImpl process = getProcess();
      process.addClassDependency(connectorClassName);
      push(deadline);
    } else {
      problems.add(new Problem("Unable to add deadline with condition " + condition + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a connector to an activity
   * @param className
   * @param event
   * @param throwingException
   * @return
   */
  public ProcessBuilder addConnector(Event event, String className, boolean throwingException) {
    Misc.checkArgsNotNull(className, event);
    if ("".equals(className.trim())) {
      problems.add(new Problem("Class name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      if (activity.isAutomatic()) {
        if (!Event.automaticOnEnter.equals(event)) {
          problems.add(new Problem("A system task accepts only a connector with an " + Event.automaticOnEnter + " event"+ " ", Problem.SEVERITY_ERROR));
        }
      } else {
        if (!Event.TASK_EVENTS.contains(event)) {
          problems.add(new Problem("A human task accepts only a connector with an event of this list: " + Event.TASK_EVENTS, Problem.SEVERITY_ERROR));
        }
      }
      ConnectorDefinitionImpl connector = new ConnectorDefinitionImpl(className);
      connector.setEvent(event);
      connector.setThrowingException(throwingException);
      activity.addConnector(connector);
      getProcess().addClassDependency(className);
      push(connector);
    } else {
      problems.add(new Problem("Unable to add a connector with class " + className + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  public ProcessBuilder addParameter(String fieldName, Object... value) {
    Misc.checkArgsNotNull(fieldName);
    if ("".equals(fieldName.trim())) {
      problems.add(new Problem("Connector field name is an empty string.", Problem.SEVERITY_ERROR));
    }
    //Object obj = peek();
    String setterName = null;
    if (fieldName.startsWith("set") || fieldName.startsWith("get")
        || fieldName.startsWith("is") || fieldName.equals("file:")
        || fieldName.equals("resource:") || fieldName.equals("bar:")) {
      setterName = fieldName;
    } else {
      StringBuilder builder = new StringBuilder("set");
      builder.append(String.valueOf(fieldName.charAt(0)).toUpperCase());
      builder.append(fieldName.substring(1));
      setterName = builder.toString();
    }
    return setParameter(setterName, value);
  }

  private ProcessBuilder setParameter(String setterName, Object... value) {
    Misc.checkArgsNotNull(setterName, value);
    if ("".equals(setterName.trim())) {
      problems.add(new Problem("Setter name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ConnectorDefinitionImpl.class);
    if (isConnector(obj)) {
      ((ConnectorDefinitionImpl)obj).addParameter(setterName, value);
    } else {
      problems.add(new Problem("Unable to add a parameter with field " + setterName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Add a filter to an activity
   * @param className
   * @return
   */
  public ProcessBuilder addFilter(String className) {
    Misc.checkArgsNotNull(className);
    if ("".equals(className.trim())) {
      problems.add(new Problem("Filter class name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      FilterDefinition filter = new ConnectorDefinitionImpl(className);
      activity.setFilter(filter);
      getProcess().addClassDependency(className);
      push(filter);
    } else {
      problems.add(new Problem("Unable to add a filter with class name " + className + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a multiInstantiation to an activity
   * @param variableId
   * @param className
   * @param parameters
   * @return
   */
  public ProcessBuilder addMultiInstanciation(String variableName, String className) {
    Misc.checkArgsNotNull(className, variableName);
    if ("".equals(variableName.trim())) {
      problems.add(new Problem("Multi instantiation variable name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if ("".equals(className.trim())) {
      problems.add(new Problem("Multi instantiation class name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      ConnectorDefinitionImpl multiInstanciation = new ConnectorDefinitionImpl(className);
      multiInstanciation.setVariableName(variableName);
      activity.setMultiInstanciation(multiInstanciation);
      getProcess().addClassDependency(className);
      push(multiInstanciation);
    } else {
      problems.add(new Problem("Unable to add a multi instantiation with variable name " + variableName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Defines that the current activity is asynchronous
   * @return
   */
  public ProcessBuilder asynchronous() {
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      activity.setAsynchronous(true);
    } else {
      problems.add(new Problem("Unable to set asynchronous on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Defines the join type of an activity
   * @param join
   * @return
   */
  public ProcessBuilder addJoinType(JoinType join) {
    Misc.checkArgsNotNull(join);
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      activity.setJoinType(join);
    } else {
      problems.add(new Problem("Unable to set join type " + join + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Defines the split type of an activity
   * @param split
   * @return
   */
  public ProcessBuilder addSplitType(SplitType split) {
    Misc.checkArgsNotNull(split);
    Object obj = peek(ActivityDefinitionImpl.class);
    if (isActivity(obj)) {
      ActivityDefinitionImpl activity = (ActivityDefinitionImpl) obj;
      activity.setSplitType(split);
    } else {
      problems.add(new Problem("Unable to set split type " + split + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a transition 
   * @param transitionName
   * @param fromActivityName
   * @param toActivityName
   * @return
   */
  public ProcessBuilder addTransition(String transitionName, String fromActivityName, String toActivityName) {
    Misc.checkArgsNotNull(transitionName, fromActivityName, toActivityName);
    if ("".equals(transitionName.trim())) {
      problems.add(new Problem("Transition name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if (fromActivityName == null) {
      problems.add(new Problem("From attribute is not defined on transition: " + transitionName, Problem.SEVERITY_ERROR));
    } else if ("".equals(fromActivityName.trim())) {
      problems.add(new Problem("Transition from activity name is an empty string.", Problem.SEVERITY_ERROR));
    }
    if (toActivityName == null) {
      problems.add(new Problem("To attribute is not defined on transition: " + transitionName, Problem.SEVERITY_ERROR));
    }
    if ("".equals(toActivityName.trim())) {
      problems.add(new Problem("Transition to activity name is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(ProcessDefinitionImpl.class);
    if (isProcess(obj)) {
      ProcessDefinitionImpl process = (ProcessDefinitionImpl) obj;
      TransitionDefinition transition = new TransitionDefinitionImpl(process.getUUID(), transitionName, fromActivityName, toActivityName);
      process.addTranistion(transition);
      push(transition);
    } else {
      problems.add(new Problem("Unable to add transition " + transitionName + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  /**
   * Adds a condition on a transition
   * @param expression
   * @param type
   * @return
   */
  public ProcessBuilder addCondition(String expression) {
    Misc.checkArgsNotNull(expression);
    if ("".equals(expression.trim())) {
      problems.add(new Problem("Condition expression is an empty string.", Problem.SEVERITY_ERROR));
    }
    Object obj = peek(TransitionDefinitionImpl.class);
    if (isTransition(obj)) {
      TransitionDefinitionImpl transition = (TransitionDefinitionImpl) obj;
      transition.setCondition(expression);
    } else {
      problems.add(new Problem("Unable to set condition " + expression + " on an object of type: " + getClass(obj), Problem.SEVERITY_ERROR));
    }
    return this;
  }

  public ProcessDefinition done() {
    checkNames();
    Misc.showProblems(problems, "process builder");
    return (ProcessDefinition) stack.get(0);
  }

  private boolean isConnector(final Object obj) {
    return obj instanceof ConnectorDefinitionImpl;
  }

  private boolean isDescriptionElement(final Object obj) {
    return obj instanceof DescriptionElementImpl;
  }
  
  private boolean isNamedElement(final Object obj) {
    return obj instanceof NamedElementImpl;
  }

  private boolean isActivity(final Object obj) {
    return obj instanceof ActivityDefinitionImpl;
  }

  private boolean isProcess(final Object obj) {
    return obj instanceof ProcessDefinitionImpl;
  }

  private boolean isParticipant(final Object obj) {
    return obj instanceof ParticipantDefinitionImpl;
  }
  
  private boolean isTransition(final Object obj) {
    return obj instanceof TransitionDefinitionImpl;
  }

  private ProcessDefinitionImpl getProcess() {
    return (ProcessDefinitionImpl) stack.get(0);  
  }
  
  private void push(final Object obj) {
    stack.push(obj);
  }

  private Object peek(Class< ? > clazz) {
    Misc.checkArgsNotNull(clazz);
    while (stack.size() > 1) {
      if (clazz.isAssignableFrom(stack.peek().getClass())) {
        return stack.peek();
      }
      stack.pop();
    }
    return stack.peek();
  }

  private String getClass(Object obj) {
    if (obj == null) {
      return null;
    }
    Class< ? > clazz = obj.getClass();
    Class< ? >[] itf = clazz.getInterfaces();
    if (itf == null || itf.length == 0) {
      return clazz.toString();
    }
    return itf[0].getName();
  }

  private void checkNames() {
    ProcessDefinitionImpl definition = (ProcessDefinitionImpl) stack.get(0);
    Set<ParticipantDefinition> groups = definition.getParticipants();
    if (groups != null) {
      List<String> groupNames = new ArrayList<String>();
      for (ParticipantDefinition group : groups) {
        String groupName = group.getName();
        if (groupNames.contains(groupName)) {
          problems.add(new Problem("Impossible to have more than one group/human with the name: " + groupName, Problem.SEVERITY_ERROR));
        } else {
          groupNames.add(groupName);
        }
      }
    }
    Set<DataFieldDefinition> processDataFields = definition.getDataFields();
    List<String> dataFieldNames = new ArrayList<String>();
    if (processDataFields != null) {
      for (DataFieldDefinition datafield : processDataFields) {
        String dataFieldName = datafield.getName();
        if (dataFieldNames.contains(dataFieldName)) {
          problems.add(new Problem("Impossible to have more than one data with the name: " + dataFieldName, Problem.SEVERITY_ERROR));
        } else {
          dataFieldNames.add(dataFieldName);
        }
      }
    }
    Set<TransitionDefinition> transitions = definition.getTransitions();
    if (transitions != null) {
      List<String> transitionNames = new ArrayList<String>();
      for (TransitionDefinition transition : transitions) {
        String transitionName = transition.getName();
        if (transitionNames.contains(transitionName)) {
          problems.add(new Problem("Impossible to have more than one transition with the name: " + transitionName, Problem.SEVERITY_ERROR));
        } else {
          transitionNames.add(transitionName);
        }
      }
    }
    Set<ActivityDefinition> activities = definition.getActivities();
    if (activities != null) {
      List<String> activityNames = new ArrayList<String>();
      for (ActivityDefinition activity : activities) {
        String activityName = activity.getName();
        if (activityNames.contains(activityName)) {
          problems.add(new Problem("Impossible to have more than one activity with the name: " + activityName, Problem.SEVERITY_ERROR));
        } else {
          activityNames.add(activityName);
        }
        Set<DataFieldDefinition> activityDataFields = activity.getDataFields();
        if (activityDataFields != null) {
          List<String> activityDataNames = new ArrayList<String>();
          for (DataFieldDefinition activityDataField : activityDataFields) {
            String activityDataName = activityDataField.getName();
            if (activityDataNames.contains(activityDataName)) {
              problems.add(new Problem("Impossible to have more than one data with name:" + activityDataName + " in the same activity", Problem.SEVERITY_ERROR));
            } else {
              activityDataNames.add(activityDataName);
            }
          }
        }
      }
    }
  }
}